@@ -1,5 +1,5 @@
 use std::str;
-use pdf::file::{File, FileOptions};
+use pdf::file::FileOptions;
 use pdf::object::*;
 use pdf::parser::{parse, ParseFlags};
 use glob::glob;
@@ -127,10 +127,12 @@ fn invalid_pdfs() {
 fn parse_objects_from_stream() {
     use pdf::object::NoResolve;
     let file = run!(FileOptions::cached().open(file_path!("xelatex.pdf")));
+    let resolver = file.resolver();
+
     // .. we know that object 13 of that file is an ObjectStream
-    let obj_stream: RcRef<ObjectStream> = run!(file.get(Ref::new(PlainRef {id: 13, gen: 0})));
+    let obj_stream: RcRef<ObjectStream> = run!(resolver.get(Ref::new(PlainRef {id: 13, gen: 0})));
     for i in 0..obj_stream.n_objects() {
-        let (data, range) = run!(obj_stream.get_object_slice(i, &file));
+        let (data, range) = run!(obj_stream.get_object_slice(i, &resolver));
         let slice = &data[range];
         println!("Object slice #{}: {}\n", i, str::from_utf8(slice).unwrap());
         run!(parse(slice, &NoResolve, ParseFlags::ANY));
